package cz.mendelu.dnaAnalyser.analyse.g4killer;

import cz.mendelu.dnaAnalyser.analyse.g4hunter.G4HunterAnalyser;
import cz.mendelu.dnaAnalyser.sequence.data.SequenceData;
import cz.mendelu.dnaAnalyser.sequence.nucleotide.Nucleotide;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.Collections;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

@Service
@Slf4j
public class G4KillerService {

    @Autowired
    private G4HunterAnalyser g4HunterAnalyser;


    /**
     * Create simple analyze for whole input. Whole input sequence is used as one G4 seqence.
     * @param sequence G4Hunter input seqence.
     * @param threshold target threshold
     * @return
     */
    public Outcome sequenceAnalyse(SequenceData sequence, float threshold, Boolean onComplementary) {
        if (onComplementary != null && onComplementary == true) {
            SequenceData supplementSequence = sequence.complementary();
            log.debug("Supplement sequence: {} --> {}", sequence.toPlain(), supplementSequence.toPlain());
            sequence = supplementSequence;
        }
        Outcome outcome = new Outcome();
        outcome.setOnComplementary(onComplementary);
        outcome.setOriginSequence(sequence.iterator().toPlain());
        outcome.setTargetThreshold(threshold);
        outcome.setOriginScore(g4HunterAnalyser.score(sequence.iterator()));

        if (outcome.getOriginScore() >= threshold) {
            SortedSet<SequenceData> mutation = mutationToGetScoreUnderThreshold(sequence, threshold);
            outcome.setMutationSequences(mutation.stream()
                    .map(SequenceData::toPlain)
                    .collect(Collectors.toList()));
            outcome.setMutationScore(g4HunterAnalyser.score(mutation.first().iterator()));
            outcome.setMutationVariants(mutation.size());
            outcome.setChangeCount(changeCount.get());
        }
        return outcome;
    }

    private final ThreadLocal<Integer> changeCount = new ThreadLocal<>();

    private SortedSet<SequenceData> mutationToGetScoreUnderThreshold(SequenceData sequence, float threshold) {
        changeCount.set(0);
        SortedSet<SequenceData> set = new TreeSet<>(Collections.singleton(sequence));
        while (threshold <= minScore(set)) {
            set = mutationCycle(set);
            changeCount.set(changeCount.get() + 1);
        }
        return set;
    }

    private float minScore(Set<SequenceData> sequenceDataSet) {
        return Double.valueOf(sequenceDataSet.stream()
                .map(SequenceData::iterator)
                .mapToDouble(g4HunterAnalyser::score)
                .min()
                .getAsDouble())
                .floatValue();
    }

    private SortedSet<SequenceData> mutationCycle(Set<SequenceData> sequenceDataSet) {
        SortedSet<SequenceData> result = new TreeSet<>();
        float min = Float.MAX_VALUE;
        for (SequenceData sequenceData: sequenceDataSet) {
            for (SequenceData mutation: mutation(sequenceData)) {
                float score = g4HunterAnalyser.score(mutation.iterator());
                if (min > score) {
                    min = score;
                    result = new TreeSet<>();
                    result.add(mutation);
                    continue;
                }

                if (min == score) {
                    result.add(mutation);
                }
            }
        }
        return result;
    }

    private Set<SequenceData> mutation(SequenceData sequenceData) {
        return IntStream.range(0, sequenceData.getLength())
                .filter(i -> sequenceData.get(i) == Nucleotide.G)
                .mapToObj(i -> sequenceData.mutation(i, Nucleotide.W))
                .collect(Collectors.toSet());
    }
}
