package cz.mendelu.dnaAnalyser.sequence.data;

import cz.mendelu.dnaAnalyser.sequence.NucleicBuffer;
import cz.mendelu.dnaAnalyser.sequence.Sequence;
import cz.mendelu.dnaAnalyser.sequence.SequenceType;
import cz.mendelu.dnaAnalyser.sequence.nucleotide.Nucleotide;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.builder.ToStringBuilder;

import java.nio.ByteBuffer;
import java.util.Objects;
import java.util.UUID;

import static java.lang.Math.abs;
import static java.lang.Math.min;
import static java.text.MessageFormat.format;

/**
 * Created by Jan Kolomazník on 30.6.17.
 */
@Slf4j
public class SequenceData implements Iterable<Nucleotide>, Comparable<SequenceData> {

    //only for testing
    private static SequenceData wrap(Sequence sequence, byte[] buffer) {
        return new SequenceData(sequence, ByteBuffer.wrap(buffer));
    }

    public static SequenceData wrap(Sequence sequence, ByteBuffer buffer) {
        return new SequenceData(sequence, buffer);
    }

    /**
     * Create sequenceData from String. This method do not store SequenceData in repository.
     *
     * @param sequence In valid string format: <code>ACTGTTTGAA</code>
     * @param cyclic   true if sequence si cyclic.
     * @param type     Type of sequence {@link SequenceType}
     * @return New not stored DataSeqence
     */
    public static SequenceData wrap(String sequence, boolean cyclic, SequenceType type) {
        ByteBuffer byteBuffer = ByteBuffer.wrap(sequence.getBytes());
        return wrap(byteBuffer, cyclic, type);
    }

    public static SequenceData wrap(ByteBuffer buffer, boolean cyclic, SequenceType type) {
        return new SequenceData(buffer, cyclic, type);
    }

    /**
     * for usage of one sequence across threads use this method to get independent instance
     */

    public SequenceData duplicate() {
        return new SequenceData(buffer, cyclic, type);
    }

    public SequenceData complementary() {
        ByteBuffer supplementBuffer = ByteBuffer.allocate(buffer.capacity());
        for(int i = 0; i < buffer.capacity(); i++) {
            supplementBuffer.put(i, this.get(i).supplement(type).toByte());
        }
        return new SequenceData(supplementBuffer, cyclic, type);
    }

    private final UUID uuid;
    private ByteBuffer buffer;
    private final boolean cyclic;
    private final SequenceType type;

    /**
     * Create in-memory sequence data not associated with any sequence.
     *
     * @param buffer ByteBuffer of nuclides
     * @param cyclic cyclic/acyclic variant
     * @param type   type of sequence
     */
    protected SequenceData(ByteBuffer buffer, boolean cyclic, SequenceType type) {
        this.buffer = buffer.asReadOnlyBuffer();
        this.cyclic = cyclic;
        this.type = type;
        this.uuid = null;
    }

    protected SequenceData(Sequence sequence, ByteBuffer buffer) {
        assert sequence != null : "Sequence is null.";
        assert buffer != null : "Buffer is null.";

        this.uuid = sequence.getId();
        this.buffer = buffer;
        this.cyclic = (sequence.getCircular() == null)
                ? false
                : sequence.getCircular();
        this.type = (sequence.getType() == null)
                ? SequenceType.DNA
                : sequence.getType();

    }

    public NucleicBuffer getBuffer(int position, int length) {
        //todo this could be improved with newer java(=>9) https://docs.oracle.com/javase/10/docs/api/java/nio/ByteBuffer.html#alignedSlice(int)
        if (position < 0 || position > getLength()) {
            throw new IllegalArgumentException(format("Position {0} i out of range", position));
        }
        return (cyclic) ? new CyclicBuffer(position, position + length) : new TrimBuffer(position, position + length);
    }

    public String toPlain() {
        return iterator().toPlain();
    }


    public int getLength() {
        return buffer.limit();
    }

    public UUID getUuid() {
        return uuid;
    }

    public SequenceType getType() {
        return type;
    }

    public boolean isCyclic() {
        return cyclic;
    }

    @Override
    public NucleicBuffer iterator() {
        return getBuffer(0, buffer.limit());
    }

    @Override
    public String toString() {
        return new ToStringBuilder(this)
                .append("sequence.uuid", getUuid())
                .append("buffer.length", getLength())
                .toString();
    }

    // is thread unsave
    public String toString(final int pos, int len) {
        if (pos > getLength()) {
            return "";
        }

        len = (pos + len > buffer.limit())
                ? buffer.limit() - pos
                : len;
        byte[] raw = new byte[len];
        buffer.position(pos); // Fix Java bug
        buffer.get(raw, 0, len);
        return new String(raw);
    }

    public Nucleotide get(int i) {
        return Nucleotide.getFromInternalFormat(buffer.get(i));
    }

    public SequenceData mutation(int i, Nucleotide nucleotide) {
        buffer.position(0); // Fix Java bug
        ByteBuffer copy = deepCopyByteBuffer(buffer);
        copy.put(i, nucleotide.toByte());
        return SequenceData.wrap(copy, cyclic, type);
    }


    private ByteBuffer deepCopyByteBuffer(final ByteBuffer original) {
        final ByteBuffer clone = (original.isDirect()) ?
                ByteBuffer.allocateDirect(original.capacity()) :
                ByteBuffer.allocate(original.capacity());
        clone.put(original.asReadOnlyBuffer());
        return clone;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        SequenceData other = (SequenceData) o;
        return this.cyclic == other.cyclic &&
                this.type == other.type &&
                Objects.equals(this.toPlain(), other.toPlain());
    }

    @Override
    public int hashCode() {
        int result = Objects.hash(cyclic, type);
        result = 31 * result + Objects.hashCode(buffer);
        return result;
    }

    @Override
    public int compareTo(SequenceData o) {
        String thisPlain = this.toPlain();
        String otherPlain = o.toPlain();
        return thisPlain.compareTo(otherPlain);
    }

    private abstract class AbstractBuffer implements NucleicBuffer {

        protected final int start, stop;
        protected int pos;

        /**
         * creates new independent instance
         */
        public AbstractBuffer(int start, int stop) {
            this.start = start;
            this.stop = stop;
            this.pos = 0;
            buffer = buffer.asReadOnlyBuffer();
        }

        @Override
        public int getAbsoluteStartPosition() {
            return start;
        }

        @Override
        public int getLength() {
            return abs(stop - start);
        }

        @Override
        public int getPosition() {
            return start + pos;
        }

        @Override
        public SequenceType getType() {
            return type;
        }

        @Override
        public boolean isCyclic() {
            return cyclic;
        }

        @Override
        public NucleicBuffer getSubBuffer(int position, int length) {
            //todo this could be improved with newer java(=>9) https://docs.oracle.com/javase/10/docs/api/java/nio/ByteBuffer.html#alignedSlice(int)
            if (position < 0 || position > getLength()) {
                throw new IllegalArgumentException(format("Position {0} i out of range", position));
            }
            return (cyclic) ? new CyclicBuffer(position, position + length) : new TrimBuffer(position, position + length);
        }

        @Override
        public Nucleotide get(int pos) {
            int absPos = absolutePosition(pos);
            if (!isAbsolutePositionInBufferRange(absPos)) {
                throw new IndexOutOfBoundsException(format("Position {0} is out of range.", pos));
            }
            return Nucleotide.getFromInternalFormat(buffer.get(absPos));
        }

        @Override
        public Nucleotide next() {
            int absPos = absolutePosition(pos++);
            if (!isAbsolutePositionInBufferRange(absPos)) {
                throw new IndexOutOfBoundsException(format("Position {0} is out of range.", pos));
            }
            return Nucleotide.getFromInternalFormat(buffer.get(absPos));
        }

        //is thread unsafe
        public String toPlain() {
            log.trace("Buffer real class is {}", buffer.getClass());
            int offset = min(start, stop);
            int length = min(abs(start - stop), abs(start - buffer.limit()));
            byte[] arr = new byte[length];
            buffer.position(offset); // Fix Java bug
            buffer.get(arr, 0, length);
            return new String(arr);
        }

        @Override
        public String toString() {
            return getClass().getSimpleName() + "("
                    + "start=" + start + ", "
                    + "stop=" + stop + ", "
                    + "pos=" + pos + ")";
        }

        protected int absolutePosition(int pos) {
            return (start < stop)
                    ? start + pos
                    : start - pos;
        }

        protected boolean isAbsolutePositionInBufferRange(int absolutePosition) {
            return (start < stop)
                    ? start <= absolutePosition && absolutePosition < stop
                    : start >= absolutePosition && absolutePosition > stop;
        }
    }

    private class TrimBuffer extends AbstractBuffer implements NucleicBuffer {

        public TrimBuffer(int start, int stop) {
            super(start, min(stop, buffer.limit()));
        }

        @Override
        public NucleicBuffer revert() {
            return new TrimBuffer(stop - 1, start - 1);
        }

        @Override
        public boolean hasNext() {
            int absPos = absolutePosition(pos);
            return isAbsolutePositionInBufferRange(absPos);
        }
    }

    private class CyclicBuffer extends AbstractBuffer implements NucleicBuffer {

        public CyclicBuffer(int start, int stop) {
            super(start, stop);
        }

        @Override
        public NucleicBuffer revert() {
            return new CyclicBuffer(stop - 1, start - 1);
        }

        @Override
        public Nucleotide get(int pos) {
            int absPos = absolutePosition(pos) % buffer.limit();
            return Nucleotide.getFromInternalFormat(buffer.get(absPos));
        }

        @Override
        public Nucleotide next() {
            int absPos = absolutePosition(pos++) % buffer.limit();
            return Nucleotide.getFromInternalFormat(buffer.get(absPos));
        }

        @Override
        public boolean hasNext() {
            int absPos = absolutePosition(pos);
            return isAbsolutePositionInBufferRange(absPos);
        }
    }




}
